class BaseOptimizer:
    def __init__(self, space, params):
        raise NotImplementedError()

    def solve(model):
        raise NotImplementedError()

    def sample_feas(num_points):
        raise NotImplementedError()
